/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef SENSE_RANGER_HH
#define SENSE_RANGER_HH

#include "FixedLenTS.hh"
#include "fSeries/ASD.hh"
#include "fSeries/PSD.hh"
#include "Histogram1.hh"
#include <string>
#include <memory>
#include <iosfwd>

class psd_estimate;
class Trend;
class MonServer;

/** The ranger class calculates the distance at which an average GW signal of
 *  a specified type can be detected above a specified S/N threshold.
 *  \author John Zweizig <john.zweizig@ligo.org>
 */
class ranger {
public:
   static double H0; ///< Hubble constant in km/s/Mpc
public:
   /** Construct a ranger instance.
    *  \brief Default constructor
    */
   ranger(void);

   /** Construct a ranger with a tool name.
    *  \brief Default constructor
    *  \param name tool name
    */
   ranger(const std::string& name);

   /** Move syntax constructor.
    *  \brief move constructor
    *  \param r Ranger instance to be moved
    */
   ranger(ranger&& r) = default;

   /** Destroy a ranger instance.
    *  \brief Destructor
    */
   ~ranger(void);

   /** Define the trend channels for this range calculator.
    *  \brief Book trend channels.
    *  \param tr Process-wide trender instance.
    */
   void bookTrends(Trend& tr);

   /** Calculate the range and red-shift using the specified data series.
    *  \brief Calculate range and red-shift.
    *  \param data Pre-conditioned h(t) time series.
    */
   void calculate(const TSeries& data);

   /** perform the requested frequency-domain calibration on the psd for the 
    *  current stride.
    *  \brief calibrate psd.
    */
   void calibrate(void);

   /** reset the las result.
    */
   void clear(void);

   /** Dump out the current parameters and status of the ranger instance
    *  to the specified output stream.
    *  \brief Dump the parameters and status.
    *  \param out standard ostream to recive status.
    */
   void display(std::ostream& out) const;

   /** Enable the red shift calculation and report range values after the red 
    *  shift calculation.
    *  \brief Enable the red shift calculation.
    *  \param rs Correct for red shift if true.
    */
   void enableRedShift(bool rs);

   /** Define the trend channels for this range calculator.
    *  \brief Book trend channels.
    *  \param tr Process-wide trender instance.
    */
   void fillTrends(Trend& tr);

   /** Get the maximum frequency to be used in the range calculation.
    *  \brief Maximum frecuency.
    *  \return Maximum frequency.
    */
   double fmax(void) const;

   /** Get the minimum frequency to be used in the range calculation.
    *  \brief Minimum frecuency.
    *  \return Minimum frequency.
    */
   double fmin(void) const;

   /** Get the signal template PSD. The specified name may be either a 
    *  pre-defined template name, e.g. "bns-1.4x1.4", or the path of a 
    *  file containing the Psd.
    *  \brief Get the signal PSD.
    *  \param template name.
    *  \return template PSD.
    */
   const containers::PSD get_template(const std::string& file) const;

   /** Set the monitor channels based on a default prefix name.
    *  \brief set monitor channels.
    *  \param prefix Prefix to use in generating channel names.
    */
   void monitorDefaultPrefix(const std::string& prefix);

   /** Set up the channels to be served by the monitor.
    *  \brief set up monitor channels.
    */
   void monitorSetup(MonServer& server);

   /** Get the monitor history length.
    *  \brief monitor history length.
    *  \return monitor history length.
    */
   Interval monitor_history_length(void) const;

   /** Extend the history series to the dpecified time. The time increment
    *  must be specified for the first call, but may be omitted thereafter.
    *  \brief Padd the history series.
    *  \param tEnd End time for the history series.
    *  \param dT   Time increment for the istory series.
    */
   void padHistory(const Time& tEnd, Interval dT);

   /** Get the psd tool type string
    *  \brief psd tool type.
    *  \return psd tool type string.
    */
   std::string psd_tool_type(void) const;

   /** Get the psd window type string
    *  \brief psd window type.
    *  \return psd window type string.
    */
   std::string psd_window_type(void) const;

   /** Get the result of the last range calculation.
    *  \brief Latest range.
    *  \return Latest rangte in Mpc. 
    */
   double range(void) const;

   /** calculate the range using the signal psd (optionally red-shifted 
    *  argument), the current calibrated noise spectrum and the significance 
    *  threshold. The differential S/N spectrum is returned as a PSD.
    *  \brief Calculate range.
    *  \param psd       Signal psd.
    *  \param integrand return integrand terms as a PSD
    *  \return Range.
    */
   double range(const containers::PSD& psd, containers::PSD& integrand) const;

   /** Calculate range averaged over the indicated time interval before the 
    *  end-time, suppressing zero samples.
    *  \brief recent range average.
    *  \param dt Averaging interval time ending with the most recent stride.
    *  \return Recent average range.
    */
   double recent(Interval dt) const;

   /** Reference to the GW template name.
    *  \brief GW template name
    *  \return GW template name string reference.
    */
   const std::string& refGWType(void) const;

   /** Return reference to the range integrand monitor channel name.
    *  \brief Integrand monitor channel reference.
    *  \return Integrand monitor channel reference.
    */
   const std::string& refIntegrandMonitor(void) const;

   /** Return reference to the range integrand frequency series
    *  \brief Range integrand psd
    *  \return Integrand psd reference.
    */
   const containers::PSD& refIntegrandSeries(void) const;

   /** Return reference to the Psd monitor channel name.
    *  \brief set the Psd monitor channel.
    *  \return Psd monitor channel name.
    */
   const std::string& refPsdMonitor(void) const;

   /** Return reference to the range trend channel name.
    *  \brief set the range trend channel.
    *  \return Range trend channel name.
    */
   const std::string& refRangeTrend(void) const;

   /** Return reference to the range monitor channel name.
    *  \brief set the range monitor channel.
    *  \return Range monitor channel name.
    */
   const std::string& refRangeMonitor(void) const;

   /** Return reference to the red shift trend channel name.
    *  \brief set the red shift trend channel.
    *  \param name Red Shift trend channel name.
    */
   const std::string& refRedShiftTrend(void) const;

   /** Return reference to the red shift monitor channel name.
    *  \brief Red shift monitor channel name reference.
    *  \return Red shift monitor channel name string reference.
    */
   const std::string& refRedShiftMonitor(void) const;

   /** Return reference to the ranger tool name.
    *  \brief Tool name.
    *  \return Tool name string reference.
    */
   const std::string& refToolName(void) const;

   /** Enable the red shift calculation and report range values after the red
    *  shift calculation.
    *  \brief Enable the red shift calculation.
    *  \param name Red Shift trend channel name.
    */
   bool redShift(void) const;

   /** Define the psd estimation tool to be used to calculate the noise
    *  spectrum.
    *  \brief Return reference to the psd estimation tool.
    *  \param est Psd estimation tool to be cloned and used.
    */
   const psd_estimate& refPSDTool(void) const;

   /** Return reference to the weight spectrum (psd) for the signal being
    *  analysed by this ranger instance. This is basically the signal
    *  amplitude squared for a source at 1Mpc (assumed z=0).
    *  \brief Get the 1Mpc signal amplitude.
    *  \param name Range trend channel name.
    */
   const containers::PSD& refWeightZ0(void) const;

   /** Reset internal history.
    *  \brief reset.
    */
   void reset(void);

   /** Set the minimum frequency used in the range calculation
    *  \brief set minimum frequency
    *  \param fmin Minimum frequency.
    */
   void setFmin(double fmin);

   /** Set the maximum frequency used in the range calculation
    *  \brief set maximum frequency
    *  \param fmax Maximum frequency.
    */
   void setFmax(double fmax);

   /** Set the GW template.
    */
   void setGWType(const std::string& name);

   /** Set the range histogram binning. The histogram is reset and the
    *  bins are set according to the parameters.
    *  \brief Set the range histogram bins.
    *  \param nBin Number of histogram bins.
    *  \param low  lower edge of the first bin.
    *  \param high upper edge of the last bin.
    */
   void setHistogramBins(int nBin, double low, double high);

   /** Set the length of the monitor time series objects.
    *  \brief Set the monitor history length.
    *  \param len Length of the monitor time series in seconds.
    */
   void setHistoryLength(Interval len);

   /** Set the integrand monitor channel name.
    *  \brief set the rpsd monitor channel.
    *  \param name Psd monitor channel name.
    */
   void setIntegrandMonitor(const std::string& name);

   /** Set the psd monitor channel name.
    *  \brief set the rpsd monitor channel.
    *  \param name Psd monitor channel name.
    */
   void setPsdMonitor(const std::string& name);

   /** Set the range histogram monitor name.
    *  \brief set the range histogram monitor channel.
    *  \param name Range histogram monitor channel name.
    */
   void setRangeHistoName(const std::string& name);

   /** Set the range monitor channel name.
    *  \brief set the range monitor channel.
    *  \param name Range monitor channel name.
    */
   void setRangeMonitor(const std::string& name);

   /** Set the range trend channel name.
    *  \brief set the range trend channel.
    *  \param name Range trend channel name.
    */
   void setRangeTrend(const std::string& name);

   /** Set the red shift trend channel name.
    *  \brief set the red shift trend channel.
    *  \param name Red Shift trend channel name.
    */
   void setRedShiftTrend(const std::string& name);

   /** Set the red shift monitor channel name.
    *  \brief set the red shift monitor channel.
    *  \param name Red shift monitor channel name.
    */
   void setRedShiftMonitor(const std::string& name);

   /** Define the psd estimation tool to be used to calculate the noise
    *  spectrum.
    *  \brief Set the psd estimation tool.
    *  \param est Psd estimation tool to be cloned and used.
    */
   void setPSDTool(const psd_estimate& est);

   /** Set the sigma threshold for the range calculation.
    *  \brief set the threshold.
    *  \param sigma Sigma threshold
    */
   void setThreshold(double sigma);

   /** Set the range integrand monitor channel name.
    *  \brief Set the integrand monitor channel.
    *  \param name Integrand monitor channel name.
    */
   void setIntegrandMonitor(const std::string& name) const;

   /** Set the weight spectrum (psd) for the signal being analysed by this
    *  ranger instance. This is basically the signal amplitude squared for a
    *  source at 1Mpc (assumed z=0).
    *  \brief Set the 1Mpc signal amplitude.
    *  \param name Range trend channel name.
    */
   void setWeightZ0(const containers::PSD& name);

   /** Get the sigma threshold.
    *  \brief Threshold
    *  \return signma threshold.
    */
   double threshold(void) const;

private:
   std::string     mToolName;
   std::string     mRangeTrend;
   std::string     mRangeMonitor;
   std::string     mRedShiftTrend;
   std::string     mRedShiftMonitor;
   std::string     mPsdMonitor;
   std::string     mIntegrandMonitor;
   std::string     mRangeHistoName;
   std::string     mGWType;
   bool            mRedShift;
   double          mSigThreshold;    ///< Threshold significance.
   double          mFmin;            ///< Min. frequency for range calculation
   double          mFmax;            ///< Max. frequency for range calculation
   std::unique_ptr<psd_estimate> mPsdTool;
   double          mWindowLength;
   Interval        mHistoryLength;   ///< Number of seconds of history
   containers::PSD mWeightZ0;        ///< Signal PSD at z=0
   double          mRange;           ///< last range result
   double          mZ;               ///< Last red-shift result
   FixedLenTS      mRangeHistory;
   FixedLenTS      mRedShiftHistory;
   containers::PSD mRawPSD;
   containers::PSD mCalibPSD;
   containers::ASD mNoiseASD;
   containers::PSD mIntegrand;
   Histogram1      mRangeHisto;
   Time            mStart;
   Time            mCurrent;
   Time            mLast;
};

//======================================  inline ranger methods
inline double
ranger::fmax(void) const {
   return mFmax;
}

inline double
ranger::fmin(void) const {
   return mFmin;
}

inline Interval
ranger::monitor_history_length(void) const {
   return mHistoryLength;
}

inline double
ranger::range(void) const {
   return mRange;
}

inline const std::string&
ranger::refGWType(void) const {
   return mGWType;
}

inline const std::string&
ranger::refRangeTrend(void) const {
   return mRangeTrend;
}

inline const std::string&
ranger::refRangeMonitor(void) const {
   return mRangeMonitor;
}

inline const std::string&
ranger::refRedShiftTrend(void) const {
   return mRedShiftTrend;
}

inline const std::string&
ranger::refRedShiftMonitor(void) const {
   return mRedShiftMonitor;
}

inline bool
ranger::redShift(void) const {
   return mRedShift;
}

inline const psd_estimate&
ranger::refPSDTool(void) const {
   return *mPsdTool;
}

inline const std::string&
ranger::refToolName(void) const {
   return mToolName;
}

inline const containers::PSD&
ranger::refWeightZ0(void) const {
   return mWeightZ0;
}

inline double
ranger::threshold(void) const {
   return mSigThreshold;
}

#endif
