/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "coherence_monitor.hh"
#include "frame_name.hh"
#include "fSeries/DFT.hh"
#include "fSeries/PSD.hh"
#include "GDSPlot.hh"
#include "window_api.hh"
#include "html/button.hh"
#include "html/document.hh"
#include "html/hline.hh"
#include "html/link.hh"
#include "html/Table2C.hh"
#include "html/text.hh"
#include "html/align.hh"
#include "html/script.hh"
#include "html/size.hh"
#include "html/writer.hh"
#include "jsonStack.hh"
#include <stdexcept>
#include <algorithm>
#include <sstream>
#include <iomanip>

using namespace containers;
using namespace html;
using namespace std;

//======================================  Generate main template.
EXECDAT(Coherence_Monitor)

inline Time
alignedEnd(const Time& t0, Interval dt) {
   if (!t0 || dt <= Interval(0)) return t0;
   Time endt(t0 + dt);
   endt -= Interval(fmod(t0.totalS(), double(dt)));
   return endt;
}

//======================================  Syntax error message function.
void
syntax(void) {
   cerr << "coherence_monitor command line syntax." << endl;
   cerr << "   coherence_monitor [<datenv-options>] <config-file>" << endl;
}

//======================================  Constructor
Coherence_Monitor::Coherence_Monitor(int argc, const char *argv[])
   : DatEnv(argc, argv), mStride(4.0), mOverlap(0.0), mAverageDt(64.0),
     mThreshold(0.2), mOscList(getDacc()), mCompression("zero-suppress-or-gzip"),
     mFrWrite("coherence", 666), mTimeAnalysed(0.0)
{
   if (argc < 2) {
      cerr << "coherence_monitor: insufficient arguments." << endl;
      finish();
      syntax();
   }

   bool cmd_err = false;
   for (size_t i=1; i<argc-1; i++) {
      string argi = argv[i];
      if (argi == "-h" || argi == "--help") {
	 cmd_err = true;
      }
      else if (isDatEnvArg(argv[i])) {
	 i++;
      }
      else {
	 cerr << "coherence_monitor: unrecognized argument '" << argi
	      << "'" << endl;
	 cmd_err = true;
      }
   }
   if (cmd_err) {
      syntax();
      finish();
      return;
   }

   //-----------------------------------  Configure the monitor
   try {
      configure(argv[argc-1]);
   } catch (exception& e) {
      cerr << "coherence_monitor: Error in configuration: " << e.what() << endl;
      finish();
      return;
   }

   //-----------------------------------  Set up the Data Accessor
   getDacc().setStride(mStride);
   getDacc().addChannel(mRefChannel);
   for (auto i=mChannels.begin(); i != mChannels.end(); i++) {
      getDacc().addChannel(i->channel());
   }
   setStrideAlignment(mStride, 0);

   //-----------------------------------  Set up OSC
   if (!mOscConfig.empty()) {
      mOscList.readConfig(mOscConfig);
   }
   
   //-----------------------------------  Set up frame writer
   if (!mFrameName.empty()) {
      mFrWrite.set_file_name(mFrameName);
      mFrWrite.set_file_alignment(mFrameLength);
      mFrWrite.setCompress(FrVectRef::get_compression_mode(mCompression));
   }

   //-----------------------------------  print status
   display(cout);
}

//======================================  Constructor
Coherence_Monitor::~Coherence_Monitor(void) {
}

//======================================  Confinguration constants
size_t ndef_bands = 21;
double band_edges[] = {1, 2, 4, 8, 16, 32, 58, 62, 90, 118, 122, 256, 500, 512,
		       1000, 1024, 1500, 1536, 2048, 3072, 4095};

const char* json_keys[] = {"reference-channel", "stride", "overlap", "threshold",
			   "frame-name", "frame-length", "compression",
			   "window-length", "window-type", "window-parameter",
			   "osc-configuration", "osc-condition", "osc-gate",
			   "measurement-bands", "mask-bands", "channels", 0};

//======================================  Read the json configuration file
void
Coherence_Monitor::configure(const std::string& file) {
   jsonStack js(file);
   
   //-----------------------------------  reference channel
   if (!js.fetch_data("reference-channel", mRefChannel)) {
      throw  runtime_error("No reference channel specified");
   }

   //-----------------------------------  get the averaging interval
   double dtemp(64.0);
   js.fetch_data("stride", dtemp);
   mAverageDt = dtemp;

   //-----------------------------------  get the window length
   dtemp = 4.0;
   js.fetch_data("window-length", dtemp);
   mStride = dtemp;
   mRefDft.set_stride(mStride);

   //-----------------------------------  Get the window parameters
   std::string stemp;
   dtemp = 0;
   js.fetch_data("window-type", stemp);
   js.fetch_data("window-parameter", dtemp);
   mRefDft.set_window(stemp, dtemp);

   //-----------------------------------  Get the overlap.
   mOverlap = 0;
   if (js.fetch_data("overlap", mOverlap)) mRefDft.set_overlap(mOverlap);

   if (!js.fetch_data("measurement-bands", mBandList)) {
      mBandList.assign(band_edges, band_edges + ndef_bands);
   }
   if (mBandList.size() < 2) {
      mBandList.clear();
      double allBand[] = {0.,8192.};
      mBandList.assign(allBand, allBand+2);
   }
   if ((js.fetch_data("mask-bands", mBandMask) % 2) != 0) {
      cerr << "coherence_monitor: Warning - masked band frequencies not paired"
	   << endl;
   }

   mThreshold = 0.20;
   js.fetch_data("threshold", mThreshold);

   //-----------------------------------   Get the coherence frame stuff
   js.fetch_data("frame-name",   mFrameName);
   if (js.fetch_data("frame-length", dtemp)) mFrameLength = Interval(dtemp);
   js.fetch_data("compression", mCompression);

   //-----------------------------------   Get a list of channel names.
   if (js.push_member("channels")) {
      jsonStack::autopop jspop(js);
      while (js.iterate()) {
	 jsonStack::autopop jschans(js);
	 string name, alias;
	 if (!js.fetch_data("name", name)) {
	    throw runtime_error("Missing coherence channel name.");
	 }
	 alias = name;
	 js.fetch_data("alias", alias);
	 coherence_channel chan(name, mRefDft);
	 chan.set_alias(alias);
	 chan.set_debug(Debug());
	 double fmax = 0;
	 if (js.fetch_data("fmax", fmax)) chan.set_fmax(fmax);
	 mChannels.push_back(chan);
      }
   }

   //-----------------------------------  Get the osc configuration.
   js.fetch_data("osc-configuration", mOscConfig);
   js.fetch_data("osc-condition",     mOscCondition);
   js.fetch_data("osc-gate",          mOscGateCond);
}

//======================================  Display DVector
string
display_dvect(const vector<double>& dv, int indent) {
   int col = indent;
   ostringstream ddv;
   ddv << "[" << fixed << setprecision(2);
   for (auto i=dv.begin(); i != dv.end(); i++) {
      int width = int(log10(*i)) + 6;
      if (col + width > 80) {
	 ddv << endl << string(indent, ' ');
	 col = indent;
      }
      ddv << " " << *i;
      if (i+1 != dv.end()) ddv << ",";
      col += width;
   }
   ddv << "]";
   return ddv.str();
}

//======================================  Process a single data stride.
void 
Coherence_Monitor::display(std::ostream& out) const {
   out << "reference channel:    " << mRefChannel        << endl;
   out << "averaging stride:     " << mAverageDt         << endl;
   out << "window type:          " << mRefDft.window_type() << endl;
   out << "window parameter:     " << mRefDft.window().parameter() << endl;
   out << "window length:        " << mRefDft.stride()   << endl;
   out << "overlap fraction:     " << mRefDft.overlap()  << endl;
   out << "OSC configuration:    " << mOscConfig         << endl;
   out << "OSC condition:        " << mOscCondition      << endl;
   out << "OSC gate condition:   " << mOscGateCond       << endl;
   out << "Frame path template:  " << mFrameName         << endl;
   out << "Frame file length:    " << mFrameLength       << endl;
   out << "Frame compression:    " << mCompression       << endl;
   out << "# coherence channels: " << mChannels.size()   << endl;
   out << "Analysis Bands:       " << display_dvect(mBandList, 22) << endl;
   out << "Band Mask:            " << display_dvect(mBandMask, 22) << endl;
   out << "Coherence channels: " << endl;
   for (auto i=mChannels.begin(); i != mChannels.end(); i++) {
      out << "   alias: " << i->alias() << "  name: " << i->channel()
	  <<" stride: " << i->stride() << " window: " << i->window_type()
	  << endl;
   }
}

//======================================  Process a single data stride.
void
Coherence_Monitor::ProcessData(void) {
   Time fillTime = getDacc().getFillTime();
   Interval dt =  getDacc().getFillStride();
   Interval pStride = mStride;
   
   //-----------------------------------  Record/check stride time
   if (!mCurrentTime || mCurrentTime < fillTime) {
      if (!mStartTime) mStartTime = fillTime;
      reset();
      mFrWrite.close();
      mCurrentTime = fillTime;
   }

   //-----------------------------------  Test osc condiiton
   Time endAvg = alignedEnd(mCurrentTime, mAverageDt);
   if (mOscConfig.empty() || mOscList.satisfied(mOscCondition)) {

      //--------------------------------  add reference channel to 
      TSeries& tsref = *getDacc().refData(mRefChannel);
      mRefDft.add(tsref);
      if (!mAverageT0) mAverageT0 = mCurrentTime;

      //--------------------------------  Add data to secondary channels
      for (auto i=mChannels.begin(); i != mChannels.end(); i++) {
	 i->add(*getDacc().refData(i->channel()));
      }

      //--------------------------------  Loop over available strides
      endAvg = alignedEnd(mAverageT0, mAverageDt);
      while (mRefDft.ready() && mRefDft.start() < endAvg) {
	 //-----------------------------  Calculate reference dft conjugate.
	 DFT ref_dft = mRefDft.nextDFT();
	 if (mRefPSD.empty()) mRefPSD  = PSD(ref_dft);
	 else                 mRefPSD += PSD(ref_dft);

	 //-----------------------------  add data to all dft automata
	 for (auto i=mChannels.begin(); i != mChannels.end(); i++) {
	    i->next_stride(ref_dft);
	 }
      }
      mTimeAnalysed += dt;

      //--------------------------------  Optional printout after stride
      if (Debug() > 3) {
	 for (auto i=mChannels.begin(); i != mChannels.end(); i++) {
	    cout << "coherence stride for channel: " << i->alias()
		 << " at GPS " << mCurrentTime.getS() << endl;
	    i->ref_coherence().get_coherence().extract_psd(0, 256).Dump(cout);
	 }
      }
   }
   
   mCurrentTime += dt;
   if (mCurrentTime >= endAvg) {
      if (!mFrameName.empty() && mAverageT0) {
	 write(mAverageT0, mCurrentTime - mAverageT0);
      }
      print();
      reset();
   }
}

//======================================  Handle a monitor data interrupt
void
Coherence_Monitor::Attention(void) {
}

//======================================  coherence band structure
struct coh_band {
   coh_band(size_t id, double f0=0) : chan_id(id) {
      reset(f0);
   }
   void bump(double fj, double df, double xj) {
      f_hi = fj + df;
      coh_avg += xj * df;
   }
   coh_band& calc(void) {
      coh_avg /= f_hi - f_lo;
      return *this;
   }
   bool empty(void) const {
      return f_lo == f_hi;
   }
   void reset(double f0) {
      f_lo = f0;
      f_hi = f0;
      coh_avg = 0;
   }
   size_t chan_id;
   double f_lo;
   double f_hi;
   double coh_avg;
};
typedef std::vector<coh_band> band_list;
typedef std::vector<band_list> summary_bands;

//======================================  Coherence band test
bool
masked_band(const coh_band& band, const vector<double>& mask) {
   size_t nMask = mask.size();
   if ((nMask % 2) == 1) nMask--;
   for (size_t i=0; i<nMask; i+=2) {
      if (band.f_lo >= mask[i] && band.f_hi <= mask[i+1]) return true;
   }
   return false;
}

//======================================  Print coherence statistics
void
Coherence_Monitor::print(void) const {
   const char* time_fmt = "%3M %3d, %Y %02H:%02N:%02S %Z (GPS %s)";

   //--------------------------------  Set up a plotter
   string mondir = "plots";
   if (getenv("DMTHTMLOUT")) mondir = getenv("DMTHTMLOUT");
   while (!mondir.empty() && mondir.back() == '/') mondir.pop_back();
   string plotdir = mondir + "/coh_plots/coh_plots-%6r/%g";
   GDSPlot pl;
   
   //--------------------------------  Create index doc, add header information
   html::document doc("Coherence Monitor Results");
   doc.header() << script("text/javascript", "sorttable.js");
   doc.header() << script("text/javascript", "collapsible.js");
   doc.lineBreak();

   //--------------------------------  Print a parameter table
   Table2C params("Global Parameters", "Parameter", "Value");
   params.addRow("Reference Channel",  mRefChannel);
   params.addRow("Window type",        mRefDft.window_type());
   params.addRow("Window length",      mStride);
   params.addRow("Overlap",            mOverlap);
   params.addRow("Averaging time",     mAverageDt);
   params.addRow("Threshold",          mThreshold);
   params.addRow("Average start time", mAverageT0,   time_fmt);
   params.addRow("Start time",         mStartTime,   time_fmt);
   params.addRow("Current Time",       mCurrentTime, time_fmt);
   params.addRow("Time analysed",      mTimeAnalysed);
   params.addRow("User CPU time",      mUserTime.get_time());
   doc << (html::div() << html::align("center") << params);
   doc.lineBreak();
   doc << hline();

   //-----------------------------------  Add some links
   doc << (html::div() << html::link("Band Summary", "band_summary.html")
	   << html::align("center"));
   doc << (html::div() << html::link("Coherence plot archive", "coh_plots")
	   << html::align("center"));
   doc << hline();

   //-----------------------------------  Add a channel table
   doc << (html::div() << (text_block("h1") << "Channel table")
	   << html::align("center"));
   table chan_tab;
   chan_tab.setClass("sortable");
   chan_tab.setBorder(true);
   chan_tab.addColumn("channel");   // col 0
   chan_tab.addColumn("nAvg");      // col 1
   chan_tab.addColumn("nTrig");     // col 2
   chan_tab.addColumn("nOK");       // col 3
   chan_tab.addColumn("Band 1");    // col 4
   chan_tab.addColumn("Coh.");      // col 5
   chan_tab.addColumn("Band 2");    // col 6
   chan_tab.addColumn("Coh.");      // col 7
   chan_tab.addColumn("Band 3");    // col 8
   chan_tab.addColumn("Coh.");      // col 9
   chan_tab.addColumn("Band 4");    // col 10
   chan_tab.addColumn("Coh.");      // col 11
   for (int k=4; k<12; k+=2) chan_tab.refHeader(k).lineBreak() << "[fmin-fmax]";

   //--------------------------------  add data to all dft automata
   size_t nChan = mChannels.size();
   size_t nBand = mBandList.size() - 1;
   summary_bands sumband(nBand);
   vector<string> ploturl(nChan);
   for (size_t inx=0; inx < nChan; inx++) {
      auto i = mChannels.begin() + inx;
      ostringstream chan_time;
      chan_time << "Channel " << i->channel() << " at " << mAverageT0.getS();

      //----------------------------- get coherence, find fmax
      PSD cohi(i->ref_coherence().get_coherence());
      double fmax = i->fmax();
      if (!fmax || cohi.getHighFreq() < fmax) fmax = cohi.getHighFreq();
      
      //-----------------------------  Get channel coherence, plot it
      pl.new_plot();
      pl.xlog(false);
      pl.ylog(false);
      pl.set_range(0, fmax, 0, 1.05);
      pl.ylabel("coherence");
      pl.title(chan_time.str());
      pl.plot(cohi.extract(0, fmax));
      frame_name fname(plotdir, i->alias() + "_COHERENCE", 0, "png");
      fname.make_dir(fname.dir_name(mAverageT0), 3);
      string plotpath = fname.file_path(mAverageT0, mAverageDt);
      pl.print(plotpath);
      ploturl[inx] = plotpath.substr(plotpath.find("coh_plots/"));
      
      //--------------------------------  Find loud bands
      band_list bands;
      double dF = cohi.getFStep();
      coh_band scan(inx, dF);
      size_t iBand = 0;
      coh_band scan_band(inx, dF);
      for (size_t j=1; j<cohi.getBin(fmax); j++) {
	 double fj = cohi.getBinF(j);

	 //-----------------------------  Save and reset band accumulator
	 while (iBand < nBand && fj >= mBandList[iBand + 1]) {
	    if (!scan_band.empty()) {
	       scan_band.calc();
	       sumband[iBand].push_back(scan_band);
	    }
	    scan_band.reset(fj);
	    iBand++;
	 }

	 //-----------------------------  Test and accumulate significant bands
	 double xj = cohi(fj).Real();
	 if (xj >= mThreshold) {
	    scan.bump(fj, dF, xj);
	    scan_band.bump(fj, dF, xj);
	 } else {
	    if (!scan.empty()) bands.push_back(scan.calc());
	    scan.reset(fj + dF);
	    if (!scan_band.empty() && iBand < nBand) {
	       sumband[iBand].push_back(scan_band.calc());
	    }
	    scan_band.reset(fj + dF);
	 }
      }
      if (!scan.empty()) bands.push_back(scan.calc());
      if (!scan_band.empty() && iBand < nBand) {
	 sumband[iBand].push_back(scan_band.calc());
      }

      //--------------------------------  sort bands in descending order
      sort(bands.begin(), bands.end(),
	   [](const coh_band& a, const coh_band& b) {
	      return a.coh_avg > b.coh_avg;
	   }
	   );

      //-------------------------------  Count masked bands
      size_t nUnmasked = 0;
      for (auto j=bands.begin(); j != bands.end(); j++) {
	 if (!masked_band(*j, mBandMask)) nUnmasked++;
      }

      //-------------------------------  Fill in channel table
      int row = chan_tab.addRow();
      chan_tab.insertData(row, 0, html::link(i->channel(), ploturl[inx]));
      chan_tab.insertData(row, 1, long(i->ref_coherence().average_count()));
      chan_tab.insertData(row, 2, long(bands.size()));
      chan_tab.insertData(row, 3, long(nUnmasked));

      int nok = 0;
      for (auto j=bands.begin(); nok < 4 && j != bands.end(); j++) {
	 if (masked_band(*j, mBandMask)) continue;
	 ostringstream band_txt;
	 band_txt << fixed << setprecision(2) << j->f_lo << "-" << j->f_hi;
	 chan_tab.insertData(row, nok*2+4, band_txt.str());
	 chan_tab.insertData(row, nok*2+5, j->coh_avg, 0, 3);
	 nok++;
      }
   }
   doc << ((html::div() << chan_tab) << html::align("center"));

   //-----------------------------------  Write out the document.
   const char* dmthtmlout = getenv("DMTHTMLOUT");
   doc.install(HTMLDATADIR, dmthtmlout);
   
   string ofile;
   if (dmthtmlout) ofile = dmthtmlout;
   if (!ofile.empty() && ofile.rfind("/") + 1 != ofile.size()) ofile += "/";

   string bsfile = ofile + "band_summary.html";
   ofile += "index.html";
   ofstream out(ofile.c_str());
   html::writer html_writer(out);
   doc.write(html_writer);

   //------------------------------------  Build the band summary document
   html::document band_sumry;
   band_sumry.header() << title("Coherence Monitor Band Summary");
   band_sumry.header() << script("text/javascript", "sorttable.js");
   band_sumry.header() << rel_link("collapse_content.css");

   band_sumry << (html::div()
		  << (text_block("H1") << "Coherence Monitor - Band Summary")
		  << html::align("center"));
   band_sumry.lineBreak();

   //-----------------------------------  Loop over non-empty bands
   for (size_t i=0; i<nBand; i++) {
      if (sumband[i].empty()) continue;

      //--------------------------------  Define table columns
      table band_tab;
      band_tab.setClass("sortable");
      band_tab.setBorder(true);
      band_tab.addColumn("alias");      // col 0
      band_tab.addColumn("Band 1");     // col 1
      band_tab.addColumn("Coh.");  // col 2
      band_tab.addColumn("Band 2");     // col 3
      band_tab.addColumn("Coh.");  // col 4
      band_tab.addColumn("Band 3");     // col 5
      band_tab.addColumn("Coh.");  // col 6
      band_tab.addColumn("Band 4");     // col 7
      band_tab.addColumn("Coh.");  // col 8
      for (int k=1; k<8; k+=2) band_tab.refHeader(k).lineBreak() << "[fmin-fmax]";

      //--------------------------------  Fill in the table rows
      int jchan = -1, jcol = 0, jrow = 0;
      int nchan = 0;
      string alias;
      for (auto j=sumband[i].begin(); j != sumband[i].end(); j++) {
	 if (j->f_lo < mBandList[0]) continue;
	 if (j->chan_id != jchan) {
	    jchan = j->chan_id;
	    alias = mChannels[jchan].alias();
	    jcol = 0;
	    nchan++;
	 }

	 if ((jcol%4) == 0) {
	    jrow = band_tab.addRow();
	    band_tab.insertData(jrow, 0, html::link(alias, ploturl[jchan]));
	 }

	 ostringstream band_txt;
	 band_txt << fixed << setprecision(2) << j->f_lo << "-" << j->f_hi;
	 band_tab.insertData(jrow, (jcol % 4) * 2 + 1, band_txt.str());
	 band_tab.insertData(jrow, (jcol % 4) * 2 + 2, j->coh_avg, 0, 3);
	 jcol++;
      }
      if (jchan < 0) continue;

      //-----------------------------------  print header
      ostringstream band;
      band << fixed << setprecision(2) << "Band "
	   << mBandList[i] << "-" << mBandList[i+1] <<" Hz"
	   << "         " << nchan << " channels with " << sumband[i].size()
	   << " total sub-bands";
      band_sumry << (html::button(band.str(), "button", "collapsible")
		     // << styleAttrib("font-size:+2;")
		     << html::align("center"));
      
      //--------------------------------  Put the table in the document
      band_sumry << (html::div("collapse_div") << band_tab) << hline();
   }
   band_sumry << script("text/javascript", "collapsible.js");
   band_sumry.install(HTMLDATADIR, dmthtmlout);

   //-----------------------------------  Write out the band summary.
   ofstream bsout(bsfile.c_str());
   html::writer bsum_writer(bsout);
   band_sumry.write(bsum_writer);
}

//======================================  Reset a coherence segent
void
Coherence_Monitor::reset(void) {
   for (auto i=mChannels.begin(); i != mChannels.end(); i++) {
      i->ref_coherence().reset_accumulators();
   }
   mRefPSD.clear();
   mAverageT0 = Time(0);
}

//======================================  Reset a coherence segent
void
Coherence_Monitor::write(const Time& t0, Interval dt) {
   mFrWrite.buildFrame(t0, dt);
   mFrWrite.addSeries(mRefChannel + "_REFPSD", mRefPSD);
   for (auto i=mChannels.begin(); i != mChannels.end(); i++) {
      mFrWrite.addSeries(i->channel() + "_CSD", i->refCSD());
      mFrWrite.addSeries(i->channel() + "_PSD", i->refPSD());
   }
   
   //-----------------------------------  Write out a frame
   mFrWrite.write();
}

//===========================================================================
//          Coherence channel
//===========================================================================

Coherence_Monitor::coherence_channel::coherence_channel(const string& name,
							const calc_dft& cd)
   : mChannel(name), mFmax(0), mDebug(0) 
{
   try {
      mCoherence.set_stride(cd.stride());
      mCoherence.set_window(cd.window());
      mCoherence.set_overlap(cd.overlap());
   } catch (std::exception& e) {
   }
}

void
Coherence_Monitor::coherence_channel::add(const TSeries& ts) {
   mCoherence.add(ts);
   if (mDebug) {
      cout << "coherence_channel: add series data to channel: " << mAlias
	   << "." << endl;
      ts.dump_header(cout);
   }
}

void
Coherence_Monitor::coherence_channel::next_stride(const containers::DFT& dft) {
   mCoherence.add(TSeries::null_tseries, dft);
}

void
Coherence_Monitor::coherence_channel::set_alias(const std::string& alias) {
   mAlias = alias;
}

void
Coherence_Monitor::coherence_channel::set_debug(int lvl) {
   mDebug = lvl;
}

void
Coherence_Monitor::coherence_channel::set_fmax(double fmax) {
   mFmax = fmax;
}
